@@ -37,26 +37,34 @@ static void fs_open(char* device){
     int flags;
     int debug = 2;
 
-    if (use_superblock && !use_blocksize)
-    	use_blocksize = 1024;
-    flags = EXT2_FLAG_JOURNAL_DEV_OK;
-
-    retval = ext2fs_open (device, flags, use_superblock, use_blocksize, unix_io_manager, &fs);
+    flags = EXT2_FLAG_JOURNAL_DEV_OK | EXT2_FLAG_SOFTSUPP_FEATURES;
+    if (use_superblock && !use_blocksize) {
+	for (use_blocksize = EXT2_MIN_BLOCK_SIZE; use_blocksize <= EXT2_MAX_BLOCK_SIZE; use_blocksize *= 2) {
+	    retval = ext2fs_open (device, flags, use_superblock, use_blocksize, unix_io_manager, &fs);
+	    if (!retval)
+		break;
+	}
+    } else
+	retval = ext2fs_open (device, flags, use_superblock, use_blocksize, unix_io_manager, &fs);
 
     if (retval) 
-        log_mesg(0, 1, 1, debug, "Couldn't find valid filesystem superblock.\n");
+	log_mesg(0, 1, 1, debug, "Couldn't find valid filesystem superblock.\n");
 
-    //if (!force && ((fs->super->s_lastcheck < fs->super->s_mtime) ||
-    if ((fs->super->s_lastcheck < fs->super->s_mtime) ||
-		(fs->super->s_state & EXT2_ERROR_FS) ||
-		((fs->super->s_state & EXT2_VALID_FS) == 0)) {
-        log_mesg(0, 1, 1, debug, "Filesystem isn't in valid state. May be it is not cleanly unmounted.\n\n");
+    ext2fs_mark_valid(fs);
+
+    if ((fs->super->s_state & EXT2_ERROR_FS) || !ext2fs_test_valid(fs))
+	log_mesg(0, 1, 1, debug, "FS contains a file system with errors\n");
+    else if ((fs->super->s_state & EXT2_VALID_FS) == 0)
+	log_mesg(0, 1, 1, debug, "FS was not cleanly unmounted\n");
+    else if ((fs->super->s_max_mnt_count > 0) && (fs->super->s_mnt_count >= (unsigned) fs->super->s_max_mnt_count)) {
+	log_mesg(0, 1, 1, debug, "FS has been mounted %u times without being checked\n");
     }
+
 }
 
 /// close device
 static void fs_close(){
-    ext2fs_close (fs);
+    ext2fs_close(fs);
 }
 
 /// get block size from super block
@@ -69,7 +77,7 @@ static unsigned long long device_size(char* device){
     int size;
     unsigned long long dev_size;
     ext2fs_get_device_size(device, EXT2_BLOCK_SIZE(fs->super), &size);
-	dev_size = (unsigned long long)(size * EXT2_BLOCK_SIZE(fs->super));
+    dev_size = (unsigned long long)(size * EXT2_BLOCK_SIZE(fs->super));
     return dev_size;
 }
 
@@ -87,59 +95,65 @@ static unsigned long long get_used_blocks(){
 extern void readbitmap(char* device, image_head image_hdr, char* bitmap){
     errcode_t retval;
     unsigned long group;
-    unsigned long long offset, current_block, block;
+    unsigned long long current_block, block;
     unsigned long long free, used, gfree, gused;
     char *block_bitmap=NULL;
-    int debug = 1;
+    int block_nbytes;
+    unsigned long long blk_itr;
+    int bg_flags = 0;
+
+    int debug = 2;
 
     log_mesg(2, 0, 0, debug, "readbitmap %i\n",bitmap);
 
     fs_open(device);
-
-    retval = ext2fs_read_bitmaps (fs); /// open extfs bitmap
+    retval = ext2fs_read_bitmaps(fs); /// open extfs bitmap
     if (retval)
-        log_mesg(0, 1, 1, debug, "Couldn't find valid filesystem bitmap.\n");
+	log_mesg(0, 1, 1, debug, "Couldn't find valid filesystem bitmap.\n");
 
+    block_nbytes = EXT2_BLOCKS_PER_GROUP(fs->super) / 8;
     if (fs->block_map)
-        block_bitmap = fs->block_map->bitmap; /// read extfs bitmap
-    
+	block_bitmap = malloc(block_nbytes);
+
     /// initial image bitmap as 1 (all block are used)
     for(block = 0; block < image_hdr.totalblock; block++)
 	bitmap[block] = 1; 
 
     free = 0;
-    used = fs->super->s_first_data_block; /// for first data block not zero(small partition) 
+    used = 0;
     current_block = 0;
+    blk_itr = fs->super->s_first_data_block;
 
     /// each group
     for (group = 0; group < fs->group_desc_count; group++) {
-	offset = fs->super->s_first_data_block;
-	//printf("offset %i\n", offset);
+
 	gfree = 0;
 	gused = 0;
-	
-        if (block_bitmap) {
-            offset += group * fs->super->s_blocks_per_group;
+
+	if (block_bitmap) {
+	    ext2fs_get_block_bitmap_range(fs->block_map, blk_itr, block_nbytes << 3, block_bitmap);
 
 	    /// each block in group
-            for (block = 0; ((block < fs->super->s_blocks_per_group) && (current_block < (image_hdr.totalblock-1))); block++) {
-		current_block = block + offset;
+	    for (block = 0; ((block < fs->super->s_blocks_per_group) && (current_block < (image_hdr.totalblock-1))); block++) {
+		current_block = block + blk_itr;
+		if (fs->super->s_feature_ro_compat & EXT4_FEATURE_RO_COMPAT_GDT_CSUM)
+		    bg_flags = fs->group_desc[group].bg_flags;
 
 		/// check block is used or not
-                if (!in_use (block_bitmap, block)) {
-                    free++;
+		if ((!in_use (block_bitmap, block)) || (bg_flags&EXT2_BG_BLOCK_UNINIT)) {
+		    free++;
 		    gfree++;
-                    bitmap[current_block] = 0;
-		    log_mesg(3, 0, 0, 1, "free block %lu at group %i\n", (current_block), group);
-                } else {
-                    used++;
+		    bitmap[current_block] = 0;
+		    log_mesg(3, 0, 0, debug, "free block %lu at group %i\n", (current_block), group);
+		} else {
+		    used++;
 		    gused++;
-                    bitmap[current_block] = 1;
-		    log_mesg(3, 0, 0, 1, "used block %lu at group %i\n", (current_block), group);
-                }
-            }
-        block_bitmap += fs->super->s_blocks_per_group / 8; /// update extfs bitmap
-        }
+		    bitmap[current_block] = 1;
+		    log_mesg(3, 0, 0, debug, "used block %lu at group %i\n", (current_block), group);
+		}
+	    }
+	    blk_itr += fs->super->s_blocks_per_group;
+	}
 	/// check free blocks in group
 	if (gfree != fs->group_desc[group].bg_free_blocks_count)
 	    log_mesg(0, 1, 1, debug, "bitmap erroe at %i group.\n", group);
